import numpy as np
from scipy.spatial import cKDTree


class DataDrivenSolver:
    """
    Data-driven solver for truss structures

    Args
    ----
    truss :
        Object defining the truss structure
    """

    def __init__(self, truss):
        self.truss = truss

    def load_material_data(self, material_data):
        """
        Load one-dimensional material data

        Args
        ----
        material_data : ndarray, shape (n_data, 2)
            Experimentally measured ``(strain, stress)`` pairs
        """
        assert material_data.shape[1] == 2
        self.material_data = material_data
        self.n_data = material_data.shape[0]
        self.data_tree = None

    def solve(
        self,
        A=1,
        U_dict={},
        F_dict={},
        n_iterations=100,
        E_num=None,
        n_neighbors=1,
        idx=None,
        save_history=False,
    ):
        """
        Solve the static equilibrium problem for the truss structure using a
        data-driven approach

        Args
        ----
        A : float or ndarray, shape (n_lines, )
            Cross section area
        U_dict : dict
            Prescribed displacement ``{point_id: (Ux, Uy), ...}``
        F_dict : dict
            Prescribed nodal force ``{point_id: (Fx, Fy), ...}``
        n_iterations : int
            Maximimum iteration for the data-driven solver
        E_num : float
            Numerical value
        n_neighbors : int
            Number of nearest local states to look for in the material data
        idx : ndarray, shape (n_lines, )
            Initial local states
        save_history : bool
            Whether also save more iteration  history

        Returns
        -------
        u : ndarray, shape (n_ddl, )
            Displacement solution
        eps : ndarray, shape (n_lines, )
            Strain
        sig : ndarray, shape (n_lines, )
            Stress
        iter_history : ndarray
            Iteration history for objective function and if asked strain/stress
        """
        # Initialize the truss solver
        if E_num is None:
            ind = np.isclose(self.material_data[:, 0], 0)
            E_secent = self.material_data[~ind, 1] / self.material_data[~ind, 0]
            E_num = E_secent.mean()
        self.sqE = np.sqrt(E_num)
        self.iter_history = None

        # Initialize local states
        if idx is None:
            idx = np.random.randint(self.n_data, size=self.truss.n_lines)
        eps_sig_ = self.material_data[idx]
        self._iter_history(append=("idx", idx))

        # Define the zeroed Dirichlet conditions
        U_dict_0 = {}
        for key in U_dict:
            U_dict_0[key] = [0 if value is not None else None for value in U_dict[key]]

        while self._iter_history() <= n_iterations:
            # Solve the 1st problem for u driven by initial stress
            sig0 = -E_num * eps_sig_[:, 0]
            if self._iter_history() == 0:
                construct_K = True
            else:
                construct_K = False
            u, eps, _ = self.truss.solve(
                A=A, E=E_num, U_dict=U_dict, sig0=sig0, construct_K=construct_K
            )

            # Solve the 2nd problem for eta driven by initial stress and applied force
            sig0 = eps_sig_[:, 1]
            _, eps_eta, _ = self.truss.solve(
                A=A, E=E_num, U_dict=U_dict_0, F_dict=F_dict, sig0=sig0
            )
            sig = eps_sig_[:, 1] + E_num * eps_eta

            # Find the nearest material data points
            eps_sig = np.hstack([eps.reshape((-1, 1)), sig.reshape((-1, 1))])
            eps_sig_idx, eps_sig_, f_obj = self._nearest_material_points(
                eps_sig, n_neighbors=n_neighbors
            )

            # Save history
            self._iter_history(append=("idx", eps_sig_idx))
            self._iter_history(append=("f_obj", f_obj))
            if save_history:
                self._iter_history(append=("eps", eps))
                self._iter_history(append=("sig", sig))

            # Check for convergence
            if np.allclose(idx, eps_sig_idx):
                self._iter_history(finalize=True)
                return u, eps_sig[:, 0], eps_sig[:, 1], self.iter_history
            else:
                idx = eps_sig_idx.copy()
        else:
            return RuntimeError(
                f"Data-driven solver not converged after {n_iterations} iterations"
            )

    def _iter_history(self, append=None, finalize=False):
        if self.iter_history is None:
            self.iter_history = {}
            for key in ["f_obj", "idx", "eps", "sig"]:
                self.iter_history[key] = []
        if finalize:
            for key in self.iter_history:
                self.iter_history[key] = np.array(self.iter_history[key])
        elif append is not None:
            key, value = append
            self.iter_history[key].append(value)
        return len(self.iter_history["f_obj"])

    def _nearest_material_points(self, eps_sig, n_neighbors=1):
        if self.data_tree is None:
            eps_sig_data = self.material_data.copy()
            eps_sig_data[:, 0] *= self.sqE
            eps_sig_data[:, 1] /= self.sqE
            self.data_tree = cKDTree(eps_sig_data)

        # Scale the strain/stress data
        eps_sig_rescaled = eps_sig.copy()
        eps_sig_rescaled[:, 0] *= self.sqE
        eps_sig_rescaled[:, 1] /= self.sqE

        dist, idx = self.data_tree.query(eps_sig_rescaled, k=n_neighbors, n_jobs=-1)
        if n_neighbors > 1:
            dist = dist.mean(axis=1)
            k_means = self.data_tree.data[idx].mean(axis=1)
            _, idx = self.data_tree.query(k_means, k=1, n_jobs=-1)

        f_obj = self.truss.integrate(dist)
        eps_sig_optimal = self.material_data[idx]
        return idx, eps_sig_optimal, f_obj



__author__ = "Tianyi Li"
__email__ = "tianyikillua@gmail.com"
__copyright__ = "Copyright (c) 2019-2020 {} <{}>".format(__author__, __email__)
__license__ = "License :: OSI Approved :: MIT License"
__version__ = "0.0.3"
__status__ = "Development Status :: 4 - Beta"



import os

from setuptools import find_packages, setup

# https://packaging.python.org/single_source_version/
base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "ddtruss", "__about__.py"), "rb") as f:
    exec(f.read(), about)


setup(
    name="ddtruss",
    version=about["__version__"],
    packages=find_packages(),
    url="https://github.com/tianyikillua/ddtruss",
    author=about["__author__"],
    author_email=about["__email__"],
    install_requires=["numpy", "scipy"],
    description="Data-driven solver for truss structures",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license=about["__license__"],
    classifiers=[
        about["__license__"],
        about["__status__"],
        # See <https://pypi.org/classifiers/> for all classifiers.
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



import numpy as np
from scipy.linalg import lu_factor, lu_solve


class Truss:
    """
    Simple static equilibrium solver for truss structures

    Args
    ----
    points : ndarray, shape (n_points, dim)
        Point coordinates with spatial dimension ``dim``
    lines : ndarray, shape (n_lines, 2)
        Connectivity of bars
    """

    def __init__(self, points, lines):
        self.points = np.asarray(points)
        self.lines = np.asarray(lines)
        assert self.lines.shape[1] == 2

        self.n_points, self.dim = self.points.shape
        self.n_lines = self.lines.shape[0]
        self.n_ddl = self.n_points * self.dim

        # Solving
        self.B = None  # gradient matrix
        self.K = None
        self.K_lu = None
        self.F = None

    def integrate(self, fun):
        """
        Integrate a function defined on bars

        Args
        ----
        fun : ndarray, shape (n_lines, ...)
            Function to be integrated
        """
        return np.sum(self.A * self.L * fun, axis=0)

    def solve(
        self, A=1, E=1, U_dict={}, F_dict={}, sig0=None, construct_K=False
    ):
        """
        Solve the static equilibrium problem for the truss structure

        Args
        ----
        A : float or ndarray, shape (n_lines, )
            Cross section area
        E : float
            Young's modulus
        U_dict : dict
            Prescribed displacement ``{point_id: (Ux, Uy), ...}``, use ``None`` for ``Ux`` or ``Uy`` when this component is not concerned
        F_dict : dict
            Prescribed nodal force ``{point_id: (Fx, Fy), ...}``
        sig0 : ndarray, shape (n_lines, )
            Initial stress
        construct_K : bool
            Whether force reconstructing and performing LU factorization of the stiffness matrix

        Returns
        -------
        u : ndarray, shape (n_ddl, )
            Displacement solution
        eps : ndarray, shape (n_lines, )
            Strain
        sig : ndarray, shape (n_lines, )
            Stress
        """
        # Geometric and material properties
        try:
            assert len(A) == self.n_lines
            self.A = A
        except TypeError:
            self.A = A * np.ones(self.n_lines)
        except AssertionError:
            raise RuntimeError("Length of A must match the number of bars")
        self.E = E

        # Stiffness matrix
        if self.K is None or construct_K:
            self._compute_elementary_quantities()
            self._construct_K()

            # Apply Dirichlet conditions to K
            if len(U_dict) > 0:
                self.K_Dirichlet = self.K.copy()
                self._apply_Dirichlet(U_dict, K=self.K_Dirichlet)
            else:
                self.K_Dirichlet = self.K

        # LU factorization
        if self.K_lu is None or construct_K:
            self.K_lu, self.K_piv = lu_factor(self.K_Dirichlet, check_finite=False)

        # Right hand side
        self._construct_F(F_dict, sig0)
        self._apply_Dirichlet(U_dict, F=self.F)

        # Solve and post-processing
        u = lu_solve((self.K_lu, self.K_piv), self.F, check_finite=False)
        eps = self._strain(u)
        return u, eps, self.E * eps

    def _global_ddl_indices(self, point_id):
        return np.array([2 * point_id, 2 * point_id + 1], dtype=int)

    def _compute_elementary_quantities(self):
        if self.B is None:
            self.L = np.zeros(self.n_lines)
            self.B = np.zeros((self.n_lines, 4))
            for i, line in enumerate(self.lines):
                coords = self.points[line]
                x, y = coords[:, 0], coords[:, 1]
                self.L[i] = np.linalg.norm(coords[1] - coords[0])
                cos = (x[1] - x[0]) / self.L[i]
                sin = (y[1] - y[0]) / self.L[i]
                self.B[i] = np.array([-cos, -sin, cos, sin]) / self.L[i]

    def _construct_K(self):
        self.K = np.zeros((self.n_ddl, self.n_ddl))
        for i, line in enumerate(self.lines):
            global_indices = np.hstack(
                [self._global_ddl_indices(line[0]), self._global_ddl_indices(line[1])]
            )
            Ke = self.A[i] * self.E * np.outer(self.B[i], self.B[i]) * self.L[i]
            self.K[global_indices[:, np.newaxis], global_indices] += Ke

    def _apply_Dirichlet(self, U_dict={}, K=None, F=None):
        for point_id, U in U_dict.items():
            assert len(U) == 2
            idx = [i for i in range(len(U)) if U[i] is not None]
            if len(idx) == 0:
                continue
            U = np.asarray(U)[idx]
            global_indices = self._global_ddl_indices(point_id)[idx]
            if K is not None:
                K[global_indices, :] = 0
                K[global_indices[:, np.newaxis], global_indices] = np.eye(len(idx))
            if F is not None:
                F[global_indices] = U

    def _construct_F(self, F_dict={}, sig0=None):
        self.F = np.zeros(self.n_ddl)

        # Prescribed nodal forces
        for point_id, F in F_dict.items():
            assert len(F) == 2
            global_indices = self._global_ddl_indices(point_id)
            self.F[global_indices] += F

        # Initial stress
        if sig0 is not None:
            assert len(sig0) == self.n_lines
            self._compute_elementary_quantities()
            for i, line in enumerate(self.lines):
                global_indices = np.hstack(
                    [
                        self._global_ddl_indices(line[0]),
                        self._global_ddl_indices(line[1]),
                    ]
                )
                Fe = -self.A[i] * sig0[i] * self.B[i] * self.L[i]
                self.F[global_indices] += Fe

    def _strain(self, u):
        eps = np.zeros(self.n_lines)
        for i, line in enumerate(self.lines):
            global_indices = np.hstack(
                [self._global_ddl_indices(line[0]), self._global_ddl_indices(line[1])]
            )
            u_ = u[global_indices]
            eps[i] = self.B[i] @ u_
        return eps

    def plot(self, ax=None, u=None, eps=None, points_id=False, lines_id=False, line_style=None):
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()

        if points_id:
            for i, point in enumerate(self.points):
                ax.text(point[0], point[1], f"{i:d}")

        if eps is not None:
            eps_absmax = np.abs(eps).max()
            norm = plt.cm.colors.Normalize(vmin=-eps_absmax, vmax=eps_absmax)
            colors = plt.cm.coolwarm(norm(eps))

        for i, line in enumerate(self.lines):
            coords = self.points[line]
            style = "--"
            color = "gray"
            if eps is not None and u is None:
                style = "-"
                color = colors[i]
            if line_style is not None:
                style = line_style
            ax.plot(coords[:, 0], coords[:, 1], style, color=color)
            if lines_id:
                centroid = np.mean(coords, axis=0)
                ax.text(centroid[0], centroid[1], f"({i:d})")

        if u is not None:
            deformed = self.points + u.reshape((-1, 2))
            for i, line in enumerate(self.lines):
                coords = deformed[line]
                color = "C0"
                if eps is not None and u is not None:
                    color = colors[i]
                ax.plot(coords[:, 0], coords[:, 1], color=color)



from .__about__ import __author__, __email__, __license__, __status__, __version__
from .solver import DataDrivenSolver
from .truss import Truss

__all__ = [
    "__author__",
    "__email__",
    "__license__",
    "__version__",
    "__status__",
    "Truss",
    "DataDrivenSolver",
]



import numpy as np
import pytest

from ddtruss import Truss, DataDrivenSolver

points = np.array([[0, 0], [1, 0], [0.5, 0.5], [2, 1]])
lines = np.array([[0, 2], [1, 2], [1, 3], [2, 3]], dtype=int)

truss = Truss(points, lines)

E = 1.962e11
A = [2e-4, 2e-4, 1e-4, 1e-4]
U_dict = {0: [0, 0], 1: [0, 0]}
F_dict = {3: [0, -9.81e3]}

u_ref = np.array(
    [0, 0, 0, 0, 2.65165043e-4, 8.83883476e-5, 3.47902545e-3, -5.60034579e-3]
)


def test_truss():
    u, *_ = truss.solve(A=A, E=E, U_dict=U_dict, F_dict=F_dict)
    assert np.allclose(u, u_ref)


@pytest.mark.parametrize(
    "n_data", [5000, 10000]
)
def test_data_driven_solver(n_data):
    ddsolver = DataDrivenSolver(truss)

    eps_max = 1.1e-3
    eps = np.linspace(-eps_max, eps_max, n_data)
    sig = E * eps
    material_data = np.hstack([eps.reshape((-1, 1)), sig.reshape((-1, 1))])
    ddsolver.load_material_data(material_data)

    u, *_ = ddsolver.solve(A=A, U_dict=U_dict, F_dict=F_dict)
    assert np.allclose(u, u_ref, rtol=1e-2)
